//#include <stdio.h>
#include <math.h>
#include "Robot.h"
#include "Powerup.h"
#include "Item.h"
#include "Map.h"


Map::Map()
:m_max_x(0)
,m_max_y(0)
,m_wraps(false)
{
}


Map::~Map()
{
	for (mapobject_v::iterator it = m_objects.begin(); it != m_objects.end(); it++)
	{
		MapObject *p = *it;
		if (p -> DeleteByMap())
		{
			delete p;
		}
	}
}


void Map::SetMax(int x,int y)
{
	m_max_x = x;
	m_max_y = y;
}


void Map::AddObject(MapObject *p,bool delete_by_map)
{
	m_objects.push_back(p);
	p -> SetDeleteByMap(delete_by_map);
	printf("New object added - new map size is %d object%s\n",m_objects.size(),(m_objects.size() == 1) ? "" : "s");
}


void Map::IncreaseTime(long usec)
{
	for (mapobject_v::iterator it = m_objects.begin(); it != m_objects.end(); it++)
	{
		MapObject *p0 = *it;
		Robot *p = dynamic_cast<Robot *>(p0);
		if (p)
		{
			p -> Tick( usec );
			p -> Call("Main");
		}
	}
}


void Map::SpawnRandom()
{
	MapObject *p = NULL;
	switch (random() % 2)
	{
	case 0: // powerup
		p = new Powerup( *this );
		break;
	case 1: // item
		p = new Item( *this );
		break;
	}
	p -> SetPos(random() % m_max_x,random() % m_max_y);
	double dist = GetNearest(p -> GetX(),p -> GetY());
	if (dist > 20)
	{
		m_objects.push_back( p );
	}
	else
	{
		delete p;
	}
}


double Map::GetNearest(double x,double y)
{
	double dist = m_max_x * m_max_y;
	for (mapobject_v::iterator it = m_objects.begin(); it != m_objects.end(); it++)
	{
		MapObject *p = *it;
		double x1 = p -> GetX() - x;
		double x2 = x - p -> GetX();
		if (x1 < m_max_x)
			x1 += m_max_x;
		if (x2 < m_max_x)
			x2 += m_max_x;
		double y1 = p -> GetY() - y;
		double y2 = y - p -> GetY();
		if (y1 < m_max_y)
			y1 += m_max_y;
		if (y2 < m_max_y)
			y2 += m_max_y;
		double xx = (x1 < x2) ? x1 : x2;
		double yy = (y1 < y2) ? y1 : y2;
		double d = sqrt(xx * xx + yy * yy);
		if (d < dist)
			dist = d;
	}
	return dist;
}


