#ifndef _ROBOT_H
#define _ROBOT_H

#include <string>
#include <map>
#include "MapObject.h"
#include "Radar.h"
#include "Turret.h"
#include "Gun.h"
#include "Shields.h"
#include "Life.h"
#include "RobotXMLFile.h"
#include "CodeTool.h"
class Map;


class Robot : public MapObject
{
public:
	Robot(Map& ,const std::string& xmlfile);
	~Robot();

	/** XML parsed ok */
	bool IsOk() { return m_config.IsOk(); }

	mapobject_t GetType() { return MAPOBJECT_ROBOT; }
	// MapObject::get_property
	long get_property(const std::string& );

	void Tick(long usec);
	Life& PartLife() { return m_life; }
	Radar& PartRadar() { return m_radar; }
	Turret& PartTurret() { return m_turret; }
	Gun& PartGun() { return m_gun; }
	Shields& PartShields() { return m_shields; }

	long get_property(int);
	void set_property(int,long);
	long call_method(int,stack_v&);

	// language
	void Call(const std::string& method);

	// inventory
	void AddObject(MapObject *);
	mapobject_v& GetInventory() { return m_inventory; } // robot.inventory

	// power
	int GetPower() { return (int)((m_power_level * 100) / m_max_power_level); } // robot.power
	int GetPowerLevel() { return (int)m_power_level; }
	int GetMaxPowerLevel() { return m_max_power_level; }
	void SetMaxPowerLevel(int x) { m_max_power_level = x; }
	bool Charging() { return GetPowerLevel() < GetMaxPowerLevel(); } // robot.charging
	int GetChargeRate() { return m_charge_rate; }
	void SetChargeRate(int x) { m_charge_rate = x; }
	bool UsePower(double units); // decrease power level, returns false and calls OnOutOfPower if failed

	// turn
	int GetCurrentHeading() { return (int)m_current_heading; }
	void SetHeading(double x) { m_set_heading = x; }
	double GetTurnRate() { return m_turn_rate; }
	void SetTurnRate(double x) { m_turn_rate = x; }

private:
	RobotXMLFile m_config;
	CodeTool *m_code;
	// parts
	Life m_life; // Robot life
	Radar m_radar;
	Turret m_turret;
	Gun m_gun;
	Shields m_shields;
	// inventory...
	mapobject_v m_inventory;
	// power: initial charge_rate 1000 units/sec
	double m_power_level;
	int m_max_power_level;
	int m_charge_rate; // units/s (1000+)
	// turn: 550 power/sec, 12 degrees/sec
	double m_current_heading;
	double m_set_heading;
	double m_turn_rate; // 12
};


#endif // _ROBOT_H
